@@ -21,7 +21,7 @@
 
 #define USAGE " dir"
 
-#define VERSION "$Id: runsv.c,v 1.11 2003/03/09 14:05:49 pape Exp $"
+#define VERSION "$Id: runsv.c,v 1.17 2004/03/28 18:17:35 pape Exp $"
 
 char *progname;
 int selfpipe[2];
@@ -65,6 +65,9 @@ void fatal(char *m) {
 void fatal2(char *m1, char *m2) {
   strerr_die6sys(111, "runsv ", dir, ": fatal: ", m1, m2, ": ");
 }
+void fatalx(char *m1, char *m2) {
+  strerr_die5x(111, "runsv ", dir, ": fatal: ", m1, m2);
+}
 void warn(char *m) {
   strerr_warn5("runsv ", dir, ": warning: ", m, ": ", &strerr_sys);
 }
@@ -305,6 +308,9 @@ int ctrl(struct svdir *s, char c) {
   case 'i': /* sig int */
     if (s->pid) kill(s->pid, SIGINT);
     break;
+  case 'q': /* sig quit */
+    if (s->pid) kill(s->pid, SIGQUIT);
+    break;
   case '1': /* sig usr1 */
     if (s->pid) kill(s->pid, SIGUSR1);
     break;
@@ -318,6 +324,8 @@ int ctrl(struct svdir *s, char c) {
 int main(int argc, char **argv) {
   struct stat s;
   int fd;
+  int r;
+  char buf[256];
 
   progname =argv[0];
   if (! argv[1] || argv[2]) usage();
@@ -367,17 +375,46 @@ int main(int argc, char **argv) {
     }
   }
 
-  mkdir("supervise", 0700);
+  if (mkdir("supervise", 0700) == -1) {
+    if ((r =readlink("supervise", buf, 256)) != -1) {
+      if (r == 256)
+	fatalx("unable to readlink ./supervise: ", "name too long");
+      buf[r] =0;
+      mkdir(buf, 0700);
+    }
+    else {
+      if ((errno != ENOENT) && (errno != EINVAL))
+	fatal("unable to readlink ./supervise");
+    }
+  }
   if ((svd[0].fdlock =open_append("supervise/lock")) == -1)
-    fatal("unable to open lock");
-  if (lock_exnb(svd[0].fdlock) == -1) fatal("unable to lock");
+    fatal("unable to open supervise/lock");
+  if (lock_exnb(svd[0].fdlock) == -1) fatal("unable to lock supervise/lock");
   coe(svd[0].fdlock);
   if (haslog) {
-    mkdir("log/supervise", 0700);
+    if (mkdir("log/supervise", 0700) == -1) {
+      if ((r =readlink("log/supervise", buf, 256)) != -1) {
+	if (r == 256)
+	  fatalx("unable to readlink ./log/supervise: ", "name too long");
+	buf[r] =0;
+	if ((fd =open_read(".")) == -1)
+	  fatal("unable to open current directory");
+	if (chdir("./log") == -1)
+	  fatal("unable to change directory to ./log"); 
+	mkdir(buf, 0700);
+	if (fchdir(fd) == -1)
+	  fatal("unable to change back to service directory");
+	close(fd);
+      }
+      else {
+	if ((errno != ENOENT) && (errno != EINVAL))
+	  fatal("unable to readlink ./log/supervise");
+      }
+    }
     if ((svd[1].fdlock =open_append("log/supervise/lock")) == -1)
-      fatal("unable to open log/lock");
+      fatal("unable to open log/supervise/lock");
     if (lock_ex(svd[1].fdlock) == -1)
-      fatal("unable to log/lock");
+      fatal("unable to lock log/supervise/lock");
     coe(svd[1].fdlock);
   }
 
@@ -416,8 +453,10 @@ int main(int argc, char **argv) {
     char ch;
 
     if (haslog)
-      if (! svd[1].pid && svd[1].want == W_UP) startservice(&svd[1]);
-    if (! svd[0].pid && svd[0].want == W_UP) startservice(&svd[0]);
+      if (! svd[1].pid && (svd[1].want == W_UP)) startservice(&svd[1]);
+    if (! svd[0].pid)
+      if ((svd[0].want == W_UP) || (svd[0].state == S_FINISH))
+	startservice(&svd[0]);
 
     x[0].fd =selfpipe[0];
     x[0].events =IOPAUSE_READ;
@@ -449,34 +488,26 @@ int main(int argc, char **argv) {
       if (child == svd[0].pid) {
 	svd[0].pid =0;
 	pidchanged =1;
+	svd[0].ctrl &=~C_TERM;
+	taia_now(&svd[0].start);
 	if (svd[0].state != S_FINISH)
 	  if ((fd =open_read("finish")) != -1) {
 	    close(fd);
 	    svd[0].state =S_FINISH;
-	    startservice(&svd[0]);
 	    update_status(&svd[0]);
 	    break;
 	  }
 	svd[0].state =S_DOWN;
-	svd[0].ctrl &=~C_TERM;
-	taia_now(&svd[0].start);
 	update_status(&svd[0]);
-	if (svd[0].want == W_UP) {
-	  startservice(&svd[0]);
-	  break;
-	}
       }
       if (haslog) {
 	if (child == svd[1].pid) {
 	  svd[1].pid =0;
 	  pidchanged =1;
 	  svd[1].state =S_DOWN;
+	  svd[1].ctrl &=~C_TERM;
 	  taia_now(&svd[1].start);
 	  update_status(&svd[1]);
-	  if (svd[1].want == W_UP) {
-	    startservice(&svd[1]);
-	    break;
-	  }
 	}
       }
     }
@@ -484,7 +515,7 @@ int main(int argc, char **argv) {
     if (haslog)
       if (read(svd[1].fdcontrol, &ch, 1) == 1) ctrl(&svd[1], ch);
 
-    if (svd[0].want == W_EXIT && svd[0].pid == 0) {
+    if (svd[0].want == W_EXIT && svd[0].state == S_DOWN) {
       if (svd[1].pid == 0) _exit(0);
       if (svd[1].want != W_EXIT) {
 	svd[1].want =W_EXIT;
